
from liteai.core import Message
from liteai.api import chat
from liteai.utils import set_logger, show_response
from loguru import logger

from tests.base import BasicTestCase


class TestQwen(BasicTestCase):
    @classmethod
    def setUpClass(cls):
        set_logger(__name__)
        logger.info("start test qwen job")

    def basic_llm(self):
        super().basic_llm("qwen-turbo")

    def test_vision_chat(self):
        question = "这张图里有什么?"
        image_path = "./data/Pikachu.png"
        model = "qwen-vl-plus"
        messages = [Message(role="user", content=question, image=image_path)]
        resp = chat(model=model, messages=messages, stream=False, temperature=0.)
        show_response(resp)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2024/08/21 18:22:06
@Author  :   ChenHao
@Description  :   
@Contact :   jerrychen1990@gmail.com
'''


from loguru import logger
from liteai.provider.api import get_provider_key
from liteai.core import ModelCard, ModelType

ZHIPU_MODELS = [
    ModelCard(name="glm-4-0520", description="glm-4-0520", provider="zhipu"),
    ModelCard(name="glm-4-air", description="glm-4-air", provider="zhipu"),
    ModelCard(name="glm-4-airx", description="glm-4-airx", provider="zhipu"),
    ModelCard(name="glm-4-flash", description="glm-4-flash", provider="zhipu"),
    ModelCard(name="glm-4v", description="glm-4v", provider="zhipu", support_vision=True),
    ModelCard(name="emohaa", description="emohaa", provider="zhipu", support_vision=True),

    ModelCard(name="embedding-2", description="embedding-2", provider="zhipu", model_type=ModelType.EMBEDDING),
    ModelCard(name="embedding-3", description="embedding-2", provider="zhipu", model_type=ModelType.EMBEDDING),
]

OPENAI_MODELS = [
    ModelCard(name="gpt-4o", description="gpt-4o", provider="openai"),
    ModelCard(name="gpt-4", description="gpt-4", provider="openai"),
    ModelCard(name="gpt-4o-mini", description="gpt-4o-mini", provider="openai")
]


MINIMAX_MODELS = [
    ModelCard(name="abab6.5s-chat", description="abab6.5s-chat", provider="minimax"),
    ModelCard(name="speech-01-turbo", description="speech-01-turbo", provider="minimax", model_type=ModelType.TTS)
]

DOUBAO_MODELS = [
    ModelCard(name="doubao-lite-4k", description="doubao-lite-4k", provider="doubao"),
]

QWEN_MODELS = [
    ModelCard(name="qwen-turbo", description="qwen-turbo", provider="qwen"),
    ModelCard(name="qwen-vl-plus", description="qwen-vl-plus", provider="qwen", support_vision=True)
]


ALL_MODELS = ZHIPU_MODELS + OPENAI_MODELS + MINIMAX_MODELS + QWEN_MODELS + DOUBAO_MODELS
ALL_MODEL_MAP = {model.name: model for model in ALL_MODELS}


def get_modelcard(model_name: str, **kwargs) -> ModelCard:
    if model_name not in ALL_MODEL_MAP:
        logger.warning(f"get not defined model name:{model_name}")
        provider_key = kwargs.pop("provider_key", None)
        if not provider_key:
            provider_key = get_provider_key(model_name=model_name)
        logger.debug(f"provider_key: {provider_key}")

        model_card = ModelCard(name=model_name, description=model_name, provider=provider_key)
        return model_card
    return ALL_MODEL_MAP[model_name]


# __all__ = ALL_MODELS + [ALL_MODEL_MAP, get_modelcard, ALL_MODELS]



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2024/08/12 19:54:53
@Author  :   ChenHao
@Description  :   工具类
@Contact :   jerrychen1990@gmail.com
'''


from typing import Any, Callable

from litellm import Field
from loguru import logger
from openai import BaseModel

from liteai.core import Parameter, ToolCall, ToolDesc


class Tool(BaseModel):
    tool_desc: ToolDesc = Field(..., description="工具描述")
    callable: Callable = Field(..., description="工具执行函数")

    def execute(self, *args, **kwargs) -> Any:
        logger.debug(f"executing tool:{self.tool_desc.name} with function:{self.callable}, with args:{args}, kwargs:{kwargs}")
        return self.callable(*args, **kwargs)


def get_current_context(time_type: str):
    import datetime
    import time
    fmt = '%Y-%m-%d %H:%M:%S' if time_type == 'datetime' else '%Y-%m-%d'
    current_time = datetime.datetime.fromtimestamp(time.time()).strftime(fmt)
    today = datetime.date.today()
    weekday = today.weekday()
    days = ["星期一", "星期二", "星期三", "星期四", "星期五", "星期六", "星期日"]

    chinese_weekday = days[weekday]
    return dict(current_time=current_time, weekday=chinese_weekday)


def on_tool_call(tool_call: ToolCall):
    logger.debug(f"handling {tool_call}")
    assert tool_call.tool_desc is not None
    if tool_call.tool_desc.is_inner:
        tool = NAME2TOOL.get(tool_call.tool_desc.name)
        if tool is None:
            logger.warning(f"tool {tool_call.tool_desc.name} not found")
        else:
            resp = tool.execute(**tool_call.parameters)
            tool_call.resp = resp
    return tool_call.resp


CurrentContextToolDesc = ToolDesc(name="current_context", description="获取当前环境，包括日期、星期几、当前时间",
                                  parameters=[Parameter(name="time_type", description="时间格式，分为两种datetime:精确到秒, date:精确到日",
                                                        type="string", required=True)],
                                  content_resp="正在帮您获取当前时间...",
                                  is_local=True, is_inner=True)
TOOL_DESCS = [CurrentContextToolDesc]
NAME2TOOL_DESC = {tool_desc.name: tool_desc for tool_desc in TOOL_DESCS}


CURRENT_CONTEXT_TOOL = Tool(tool_desc=CurrentContextToolDesc, callable=get_current_context)

TOOLS = [CURRENT_CONTEXT_TOOL]
NAME2TOOL = {tool.tool_desc.name: tool for tool in TOOLS}


__ALL__ = ["TOOL_DESCS", "TOOLS", "NAME2TOOL_DESC", "NAME2TOOL"]
__ALL__.extend(TOOL_DESCS)
__ALL__.extend(TOOLS)

if __name__ == "__main__":
    resp = CURRENT_CONTEXT_TOOL.execute(time_type="datetime")
    logger.info(resp)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2024/07/31 15:50:13
@Author  :   ChenHao
@Description  :   
@Contact :   jerrychen1990@gmail.com
'''

from abc import abstractmethod
import os
from typing import Any, List, Tuple
from loguru import logger
from typing import Any, List
from liteai.core import Message, ModelCard, ModelResponse, ToolCall, ToolDesc, Voice
from liteai.utils import truncate_dict_strings
from snippets import jdumps, retry, multi_thread


class BaseProvider:
    key: str = None
    allow_kwargs = None
    api_key_env = None

    def __init__(self, api_key: str = None):
        if api_key:
            self.api_key = api_key
        else:
            self.api_key = os.environ.get(self.api_key_env)
        # logger.debug(f"{self.api_key=}")
        if not self.api_key and not getattr(self, "base_url"):
            raise ValueError(f"api_key is required or set {self.api_key_env} in environment variables or set base_url")

    def pre_process(self, model: ModelCard, messages: List[Message], tools: List[ToolDesc],
                    tool_calls: List[ToolCall], stream: str, **kwargs) -> Tuple[List[dict], List[dict], dict]:
        new_kwargs = dict()
        ignore_kwargs = dict()
        # logger.debug(f"{self.allow_kwargs=}")
        if self.allow_kwargs:
            for k, v in kwargs.items():
                if k in self.allow_kwargs:
                    new_kwargs[k] = v
                else:
                    ignore_kwargs[k] = v
        else:
            new_kwargs = dict(**kwargs)

        # logger.debug(f"{new_kwargs=}")
        # logger.debug(f"{ignore_kwargs=}")
        if ignore_kwargs:
            logger.warning(f"ignoring {len(ignore_kwargs)} unknown kwargs: {ignore_kwargs}")

        messages = [message.model_dump(exclude_none=True) for message in messages]

        # logger.debug(f"{messages=}")

        if not model.support_system:
            self._handle_system(model, messages, **kwargs)
        # logger.debug(f"{messages=}")

        tools = [tool.model_dump(exclude_none=True) for tool in tools]
        for tool_call in tool_calls:
            if tool_call.resp:
                tool_call_message = dict(role="tool", content=jdumps(tool_call.resp), tool_call_id=tool_call.tool_call_id)
                logger.debug(f"adding tool_call_message: {tool_call_message}")
                messages.append(tool_call_message)

        return messages, tools, new_kwargs

    def _handle_system(self, model: ModelCard, messages: List[dict], **kwargs) -> List[dict]:
        system = None
        last_user_message = None
        for message in messages:
            if message["role"] == "system":
                system = message
            if message["role"] == "user":
                last_user_message = message
        if system and last_user_message:
            logger.warning(f"model:{model.name} not support system, merge system message to last user message")
            messages.remove(system)
            if kwargs.get("handle_system", True):
                last_user_message["content"] = system["content"] + "\n" + last_user_message["content"]
        return messages

    @abstractmethod
    def post_process(self, response, **kwargs) -> ModelResponse:
        raise NotImplementedError

    @abstractmethod
    def post_process_stream(self, response) -> ModelResponse:
        raise NotImplementedError

    @abstractmethod
    def _inner_complete_(self, model: str, messages: List[dict], tools: List[ToolDesc], stream: bool, **kwargs) -> Any:
        raise NotImplementedError

    def complete(self, model: ModelCard, messages: List[Message], stream: bool,
                 tools: List[ToolDesc] = [], tool_calls: List[ToolCall] = [], **kwargs) -> ModelResponse:
        messages, dict_tools, kwargs = self.pre_process(model=model, messages=messages, tools=tools, stream=stream, tool_calls=tool_calls, ** kwargs)

        self.show_calling_info(messages, dict_tools, model, stream, **kwargs)
        response = self._inner_complete_(model.name, messages, stream=stream, tools=dict_tools, **kwargs)
        if stream:
            resp = self.post_process_stream(response, **kwargs)
        else:
            resp = self.post_process(response, **kwargs)
        resp = self.on_tool_call(resp, tools)
        return resp

    def on_tool_call(self, response: ModelResponse, tools: List[ToolDesc]) -> ModelResponse:
        tool_calls = []
        tools2desc = {tool.name: tool for tool in tools}
        tool_content = ""
        logger.debug(f"on tool call for {len(response.tool_calls)} tools")
        for tool_call in response.tool_calls:
            if tool_call.name not in tools2desc:
                continue
            tool_desc: ToolDesc = tools2desc[tool_call.name]
            tool_call.tool_desc = tool_desc
            if tool_desc.content_resp:
                tool_content = tool_desc.content_resp
            # if tool_desc.is_inner:
                # logger.debug(f"calling inner tool {tool_call.name}")
                # tool_call.resp = tool_desc.func(**tool_call.parameters)
            tool_calls.append(tool_call)
        response.tool_calls = tool_calls
        logger.debug(f"get {len(tool_calls)} valid tool calls")
        if tool_content:
            logger.debug(f"set tool content {tool_content} to response")
            response.content = tool_content if isinstance(response.content, str) else (tool_content)
        return response

    def show_calling_info(self, messages: List[Message], tools: List[dict], model: ModelCard, stream: bool, **kwargs):
        show_message = messages
        show_message = truncate_dict_strings(messages, 50, key_pattern=["url"])
        message_str = ""
        for idx, message in enumerate(show_message, start=1):
            message_str += f"[{idx}].<{message['role']}>\n{message['content']}\n"

        calling_detail = f"calling {self.key} api with {model.name=}, {stream = }\nmessages=\n{message_str}"

        if tools:
            calling_detail += f"\ntools={jdumps(tools)}"
        if hasattr(self, "base_url"):
            calling_detail += f"\nbase_url = {getattr(self, 'base_url')}"
        calling_detail += f"\nkwargs = {jdumps(kwargs)}"
        logger.debug(calling_detail)

    def tts(self, text: str, model: ModelCard, stream: bool, **kwargs) -> Voice:
        raise Exception(f"provider {self.__class__.__name__} not support tts!")

    def embedding(self, texts: str | List[str], model: ModelCard, batch_size=8, **kwargs) -> List[List[float]] | List[float]:
        batch_func = multi_thread(work_num=batch_size, return_list=True)(self._embedding_single)
        return batch_func(data=texts, model=model, **kwargs)

    def _embedding_single_try(self, text: str, model: str, **kwargs) -> List[float]:
        raise Exception(f"provider {self.__class__.__name__} not support embedding!")

    def _embedding_single(self, text: str, model: ModelCard, retry_num=2, wait_time=(1, 2), **kwargs) -> List[float]:
        if retry_num:
            attempt = retry(retry_num=retry_num, wait_time=wait_time)(self._embedding_single_try)
        else:
            attempt = self._embedding_single_try
        return attempt(text=text, model=model.name, **kwargs)





from liteai.utils import set_logger
from loguru import logger

from tests.base import BasicTestCase


class TestTGI(BasicTestCase):

    @classmethod
    def setUpClass(cls):
        set_logger(__name__)
        logger.info("start test tgi job")

    def basic_llm(self):
        super().basic_llm(model="tgi_glm3_xz",
                          base_url="http://hz-model.bigmodel.cn/servyou-api")



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2024/06/25 17:44:35
@Author  :   ChenHao
@Description  :   
@Contact :   jerrychen1990@gmail.com
'''
import base64
from io import BytesIO
import os
import sys
from typing import Iterable, List
from loguru import logger
from PIL import Image
import numpy as np
from liteai.core import ModelResponse, ToolCall
from snippets import batchify
from snippets.utils import jdumps


def show_response(response: ModelResponse, batch_size=10):
    content = response.content
    if response.tool_calls:
        for tool_call in response.tool_calls:
            logger.info(f"tool_call: {jdumps(tool_call.model_dump())}")

    if isinstance(content, str):
        logger.info("content:\n"+content)
        return content
    else:
        logger.info("stream content:")
        acc = ""
        for item in batchify(content, batch_size):
            chunk = "".join(item)
            logger.info(chunk)
            acc += chunk
        return acc


def show_embeds(embds: List[List[float]] | List[float], sample_num=2, sample_size=4):
    if isinstance(embds[0], list):
        logger.info(f"embd dim:{len(embds[0])}")
        for i, embd in enumerate(embds[:sample_num]):
            logger.info(f"embd[{i}]: {embd[:sample_size]}")
    else:
        logger.info(f"embd dim:{len(embds)}")
        logger.info(f"embd: {embds[:sample_size]}")


def set_logger(module):
    print(f"setting logger for {module=}")
    if 0 in logger._core.handlers:
        logger.remove(0)
    UNILLM_ENV = os.environ.get("UNILLM_ENV", "DEV")

    dev_fmt = "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> [<level>{level: <8}</level>] - <cyan>{file}</cyan>:<cyan>{line}</cyan>[<cyan>{name}</cyan>:<cyan>{function}</cyan>] - <level>{message}</level>"
    if UNILLM_ENV.upper() == "DEV":
        logger.add(sys.stdout, level="DEBUG", filter=lambda r: module in r["name"], format=dev_fmt, enqueue=True, colorize=True)


def image2base64(image_path):
    # 打开图像文件
    with Image.open(image_path) as image:
        # 创建一个字节流对象
        buffered = BytesIO()
        # 将图像保存到字节流对象中
        image.save(buffered, format="PNG")
        # 获取字节流中的字节数据
        img_bytes = buffered.getvalue()
        # 将字节数据编码为Base64字符串
        img_base64 = base64.b64encode(img_bytes).decode("utf-8")
    return img_base64


def truncate_dict_strings(data: dict, max_length: int, key_pattern=None) -> dict:
    def truncate_string(s):
        return s if len(s) <= max_length else s[:max_length//2] + '...' + s[-max_length//2:]

    def process_item(key, item):
        if isinstance(item, dict):
            return {k: process_item(k, v) for k, v in item.items()}
        elif isinstance(item, list):
            return [process_item(key, i) for i in item]
        elif isinstance(item, str) and key in key_pattern:
            return truncate_string(item)
        else:
            return item

    return process_item(None, data)


def get_text_chunk(chunk):
    choices = chunk.choices
    if choices:
        choice = choices[0]
        # logger.debug(f"{choice=}")
        if choice.delta and choice.delta.content:
            delta_content = choice.delta.content
            # logger.info(f"{delta_content}")
            return delta_content


def extract_tool_calls(chunks: Iterable):
    tool_calls = []
    for chunk in chunks:
        # logger.debug(f"{chunk=}")
        if delta := chunk.choices[0].delta:
            if delta.tool_calls:
                for tc in delta.tool_calls:
                    tool_call = ToolCall(name=tc.function.name, parameters=eval(tc.function.arguments), tool_call_id=tc.id)
                    tool_calls.append(tool_call)
        break
    return tool_calls


def acc_chunks(acc):
    resp_msg = "".join(acc).strip()
    logger.debug(f"model generate answer:{resp_msg}")


def get_embd_similarity(embd1: List[float], embd2: List[float]):
    embd1 = np.array(embd1)
    embd2 = np.array(embd2)
    sim = np.dot(embd1, embd2) / (np.linalg.norm(embd1) * np.linalg.norm(embd2))
    return sim


if __name__ == "__main__":
    data = {
        'name': 'Pikachu',
        'description': 'Pikachu is an Electric-type Pokémon introduced in Generation I.',
        'abilities': ['Static', 'Lightning Rod']
    }
    print(truncate_dict_strings(data, 20))



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2024/07/29 11:20:33
@Author  :   ChenHao
@Description  :   配置信息
@Contact :   jerrychen1990@gmail.com
'''


import os

LITEAI_ENV = os.environ.get("LITEAI_ENV", "dev")

LITEAI_HOME = os.environ.get("LITEAI_HOME", os.path.dirname(os.path.abspath(__file__)))
LOG_HOME = os.environ.get("LOG_HOME", os.path.join(LITEAI_HOME, "logs"))

ARK_ENDPOINT_MAP = {
    "doubao-lite-4k": "ep-20240805191433-4jh54"
}



DEFAULT_VOICE_CHUNK_SIZE = 4096 * 10
MIN_PLAY_VOICE_SIZE = 8192 * 10
MAX_PLAY_SECONDS = 60 * 60



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2024/08/06 10:05:51
@Author  :   ChenHao
@Description  :   
@Contact :   jerrychen1990@gmail.com
'''


import json
import os
from typing import List
from litellm import Tuple
from loguru import logger
import requests
from liteai.core import Message, ModelCard, ToolDesc, Voice
from liteai.provider.open_ai import OpenAIProvider
from liteai.voice import build_voice
from snippets import jdumps


class MinimaxProvider(OpenAIProvider):
    key: str = "minimax"
    allow_kwargs = {"do_sample", "stream", "temperature", "top_p", "max_tokens"}
    api_key_env = "MINIMAX_API_KEY"

    def __init__(self, api_key: str = None, **kwargs):
        super().__init__(api_key=api_key, base_url="https://api.minimax.chat/v1")

    def pre_process(self, model: ModelCard, messages: List[Message], tools: List[ToolDesc], stream: bool, **kwargs) -> Tuple[List[dict], dict]:
        if kwargs.get("temperature") == 0.:
            logger.debug(f"provider {self.key} not support temperature=0, setting temperature to 0.0001")
            kwargs["temperature"] = 0.0001
        messages, tools, kwargs = super().pre_process(model=model, messages=messages, tools=tools, stream=stream, **kwargs)
        return messages, tools, kwargs

    def tts(self, text: str, model: ModelCard, version="t2a_v2", tgt_path: str = None,
            append=False, stream=False, voice_id="tianxin_xiaoling", speed=1, pitch=0) -> Voice:

        group_id = os.environ["MINIMAX_GROUP_ID"]
        url = f"https://api.minimax.chat/v1/{version}?GroupId={group_id}"

        payload = {
            "model": model.name,
            "text": text,
            "stream": stream,
            "voice_setting": {
                "voice_id": voice_id,
                "speed": speed,
                "vol": 1,
                "pitch": pitch
            },
            "audio_setting": {
                "sample_rate": 32000,
                "bitrate": 128000,
                "format": "mp3",
                "channel": 1
            }
        }
        headers = {
            'Authorization': f'Bearer {self.api_key}',
            'Content-Type': 'application/json'
        }
        logger.debug(f"calling minimax tts url:{url} with payload:\n{jdumps(payload)}")
        response = requests.request("POST", url, headers=headers, stream=stream, json=payload)
        logger.debug(f"{response.status_code=}")
        response.raise_for_status()
        if not stream:
            data = response.json()
            if "data" not in data:
                logger.exception(data)
                raise Exception(data)
            byte_stream = bytes.fromhex(data['data']['audio'])
        else:
            def _get_bytes(chunk):
                if chunk:
                    if chunk[:5] == b'data:':
                        data = json.loads(chunk[5:])
                        if "data" in data and "extra_info" not in data:
                            if "audio" in data["data"]:
                                audio = data["data"]['audio']
                                return bytes.fromhex(audio)
            byte_stream = (e for e in (_get_bytes(chunk) for chunk in response.raw) if e)
        voice = build_voice(byte_stream=byte_stream, file_path=tgt_path)
        return voice


if __name__ == "__main__":
    provider = MinimaxProvider()
    messages = [Message(role="user", content="你好")]
    resp = provider.complete(messages=messages, model="abab6.5s-chat", stream=False)
    print(resp.content)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2024/06/25 15:54:25
@Author  :   ChenHao
@Description  :
@Contact :   jerrychen1990@gmail.com
'''


from functools import wraps

from loguru import logger


from liteai.core import *
from liteai.modelcard import ALL_MODELS, get_modelcard
from liteai.provider.api import get_provider
from typing import List

from liteai.provider.base import BaseProvider
from liteai.tool import on_tool_call
from snippets import ChangeLogLevelContext


def can_set_level(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        log_level = kwargs.pop("log_level", None)
        # logger.debug(f"set log level: {log_level}")
        if log_level:
            with ChangeLogLevelContext(module_name="liteai", sink_type="stdout", level=log_level):
                return func(*args, **kwargs)
        else:
            return func(*args, **kwargs)
    return wrapper


@can_set_level
def chat(model: str | ModelCard, messages: List[dict | Message] | str, tool_calls: List[ToolCall] = [],
         provider: str = None, api_key: str = None, base_url: str = None, tools: List[ToolDesc] = [],
         stream=False, temperature=0.7, top_p=.7,  **kwargs) -> ModelResponse:
    if isinstance(model, str):
        model: ModelCard = get_modelcard(model_name=model, provider_key=provider)

    provider: BaseProvider = get_provider(model=model, api_key=api_key, base_url=base_url)

    if isinstance(messages, str):
        messages = [Message(content=messages, role="user")]
    else:
        messages = [Message(**m) if isinstance(m, dict) else m for m in messages]

    response = provider.complete(messages=messages, model=model, stream=stream, tools=tools, tool_calls=tool_calls,
                                 temperature=temperature, top_p=top_p, **kwargs)
    return response


@can_set_level
def tts(text: str | Iterable[str], model: str | ModelCard,
        provider: str = None, api_key: str = None,
        tgt_path: str = None, stream=True, **kwargs) -> Voice:
    if isinstance(model, str):
        model: ModelCard = get_modelcard(model_name=model, provider_key=provider)
    provider: BaseProvider = get_provider(model=model, api_key=api_key)
    voice = provider.tts(text=text, model=model, stream=stream, tgt_path=tgt_path, **kwargs)
    return voice


@can_set_level
def embedding(texts: str | List[str], model: str | ModelCard,
              provider: str = None, api_key: str = None, norm=True, batch_size=8, **kwargs) -> List[List[float]] | List[float]:
    if isinstance(model, str):
        model: ModelCard = get_modelcard(model_name=model, provider_key=provider)
    provider: BaseProvider = get_provider(model=model, api_key=api_key)
    voice = provider.embedding(texts=texts, model=model, norm=norm, batch_size=batch_size, **kwargs)
    return voice


@can_set_level
def agent_chat(model: str | ModelCard, messages: List[dict | Message] | str, provider: str = None,
               api_key: str = None, base_url: str = None, tools: List[ToolDesc] = [], max_iter=3,
               stream=False, temperature=0.7, top_p=.7,  **kwargs) -> ModelResponse:
    tool_calls = []

    idx = 0
    while idx < max_iter:
        logger.info(f"agent chat iteration:{idx+1}")
        resp = chat(model=model, messages=messages, provider=provider, api_key=api_key, base_url=base_url, tools=tools, tool_calls=tool_calls,
                    stream=stream, temperature=temperature, top_p=top_p, **kwargs)
        tool_calls = resp.tool_calls
        for tool_call in tool_calls:
            on_tool_call(tool_call)
        tool_calls = [e for e in tool_calls if e.tool_desc.is_inner and e.resp]
        logger.info(f"get {len(tool_calls)} inner tool calls")
        if not tool_calls:
            return resp
        idx += 1
    return resp


def list_models() -> List[dict]:
    return [model.model_dump() for model in ALL_MODELS]



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2024/06/28 10:44:41
@Author  :   ChenHao
@Description  :
@Contact :   jerrychen1990@gmail.com
'''

from http import HTTPStatus
from typing import Any, List, Tuple

from loguru import logger

from liteai.core import ModelResponse, Message, ToolDesc, Usage
from liteai.provider.base import BaseProvider

import dashscope
from dashscope.api_entities.dashscope_response import GenerationResponse, MultiModalConversationResponse
from snippets import add_callback2gen

from liteai.utils import acc_chunks


class QwenProvider(BaseProvider):
    key: str = "qwen"
    allow_kwargs = {"stream", "temperature", "top_p", "max_tokens"}
    api_key_env = "DASHSCOPE_API_KEY"

    def __init__(self, api_key: str = None, **kwargs):
        super().__init__(api_key=api_key)
        dashscope.api_key = self.api_key

    def pre_process(self, model: str, messages: List[Message], tools: List[ToolDesc], stream: bool, **kwargs) -> Tuple[List[dict], dict]:
        messages, tools, kwargs = super().pre_process(model=model, messages=messages, tools=tools, stream=stream, **kwargs)
        if stream:
            kwargs["incremental_output"] = True
        for message in messages:
            # logger.debug(f"{message=}")
            if message.get("image"):
                message["content"] = [dict(text=message["content"]),
                                      dict(image=message["image"])]
                del message["image"]
        return messages, tools, kwargs

    def _inner_complete_(self, model, messages: List[dict], stream: bool, ** kwargs) -> Any:
        if "vl" in model:
            response = dashscope.MultiModalConversation.call(model=model,
                                                             messages=messages,
                                                             **kwargs)
        else:
            response = dashscope.Generation.call(model=model,
                                                 messages=messages,
                                                 stream=stream,
                                                 result_format="message",
                                                 **kwargs)
        return response

    def post_process(self, response: dict, **kwargs) -> ModelResponse:
        logger.debug(f"{response=}, {type(response)=}")
        if response.status_code == HTTPStatus.OK:
            if isinstance(response, GenerationResponse):
                content = response.output.choices[0].message.content
            elif isinstance(response, MultiModalConversationResponse):
                content = response.output.choices[0].message.content[0]["text"]
            usage = Usage(prompt_tokens=response.usage.input_tokens, completion_tokens=response.usage.output_tokens)
            return ModelResponse(content=content, usage=usage)
        else:
            raise Exception(f"{response.status_code} {response.message}")

    def post_process_stream(self, response, **kwargs) -> ModelResponse:
        def _handel_chunk(chunk):

            if chunk.status_code == HTTPStatus.OK:
                choices = chunk.output.choices
                if choices:
                    choice = choices[0]
                    if choice.message.content:
                        delta_content = choice.message.content
                        return delta_content

        gen = (e for e in (_handel_chunk(chunk) for chunk in response) if e)
        gen = add_callback2gen(gen, acc_chunks)
        return ModelResponse(content=gen)


if __name__ == "__main__":
    provider = QwenProvider()
    messages = [Message(role="user", content="你好")]
    resp = provider.complete(messages=messages, model="qwen-turbo", stream=False)
    print(resp.content)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2024/06/25 16:39:05
@Author  :   ChenHao
@Description  :   
@Contact :   jerrychen1990@gmail.com
'''
import itertools
import json
from typing import Any, List, Tuple

from loguru import logger
import numpy as np


from liteai.core import ModelCard, ModelResponse, Message, ToolCall, ToolDesc, Usage
from zhipuai import ZhipuAI
from liteai.provider.base import BaseProvider
from snippets import add_callback2gen

from liteai.utils import extract_tool_calls, get_text_chunk, image2base64, acc_chunks


def build_tool_calls(tool_calls) -> List[ToolCall]:
    # logger.debug(f"tool_calls: {tool_calls}")
    if tool_calls is None:
        return []
    rs = []
    for tool_call in tool_calls:
        parameters = json.loads(tool_call.function.arguments)

        tmp = ToolCall(tool_call_id=tool_call.id, name=tool_call.function.name, parameters=parameters)
        rs.append(tmp)
    # logger.debug(f"tool_calls: {rs}")
    return rs


class ZhipuProvider(BaseProvider):
    key: str = "zhipu"
    allow_kwargs = {"do_sample", "stream", "temperature", "top_p", "max_tokens", "meta"}
    api_key_env = "ZHIPUAI_API_KEY"

    def __init__(self, api_key: str = None, **kwargs):
        super().__init__(api_key=api_key)
        self.client = ZhipuAI(api_key=self.api_key)

    def pre_process(self, model: ModelCard, messages: List[Message], tools: List[ToolDesc], stream: bool, **kwargs) -> Tuple[List[dict], List[dict], dict]:
        if kwargs.get("temperature") == 0.:
            del kwargs["temperature"]
            kwargs["do_sample"] = False
        messages, _, kwargs = super().pre_process(model=model, messages=messages, tools=tools, stream=stream, **kwargs)
        # logger.debug(f"{messages=}")
        for message in messages:
            # logger.debug(f"{message=}")
            if message.get("image"):
                base64 = image2base64(message["image"])
                message["content"] = [dict(type="text", text=message["content"]),
                                      dict(type="image_url", image_url=dict(url=base64))]
                del message["image"]
        tools = [self.tool2zhipu_tool(tool) for tool in tools]
        return messages, tools, kwargs

    @classmethod
    def tool2zhipu_tool(cls, tool: ToolDesc) -> dict:
        properties = {p.name: dict(description=p.description, type=p.type) for p in tool.parameters}
        required = [p.name for p in tool.parameters if p.required]
        parameters = dict(type="object", properties=properties, required=required)

        rs = dict(type="function", function=dict(name=tool.name, description=tool.description, parameters=parameters))
        return rs

    def _inner_complete_(self, model, messages: List[dict], stream: bool, tools: List[dict], **kwargs) -> Any:
        response = self.client.chat.completions.create(
            model=model,
            messages=messages,
            stream=stream,
            tools=tools,
            **kwargs
        )
        return response

    def post_process(self, response, **kwargs) -> ModelResponse:
        logger.debug(f"{response=}")
        content = response.choices[0].message.content
        tool_calls = build_tool_calls(response.choices[0].message.tool_calls)
        usage = Usage(**response.usage.model_dump())
        return ModelResponse(content=content, usage=usage, tool_calls=tool_calls)

    def post_process_stream(self, response, **kwargs) -> ModelResponse:
        toll_call_response, response = itertools.tee(response)
        tool_calls = extract_tool_calls(toll_call_response)
        logger.debug(f"{tool_calls=}")
        gen = (e for e in (get_text_chunk(chunk) for chunk in response) if e)
        gen = add_callback2gen(gen, acc_chunks)
        return ModelResponse(content=gen, tool_calls=tool_calls)

    def _embedding_single_try(self, text: str, model: str, norm=True, **kwargs) -> List[float]:
        try:
            resp = self.client.embeddings.create(
                model=model,
                input=text,
                **kwargs
            )
            embedding = resp.data[0].embedding
            if norm:
                embedding = embedding / np.linalg.norm(embedding, 2)
                embedding = embedding.tolist()
            return embedding
        except Exception as e:
            text_sample = text[:20]
            msg = f"calling zhipu api embedding with get error: {e}, with {text_sample=}, {len(text)=}, {model=}"
            raise Exception(msg)


if __name__ == "__main__":
    provider = ZhipuProvider()
    messages = [Message(role="user", content="你好")]
    resp = provider.complete(messages=messages, model="glm-3-turbo", stream=False)
    print(resp.content)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2024/08/06 14:24:26
@Author  :   ChenHao
@Description  :   语音处理相关函数
@Contact :   jerrychen1990@gmail.com
'''
import io
import itertools
import os
import threading
import time
from typing import Iterable
from loguru import logger
from liteai.config import DEFAULT_VOICE_CHUNK_SIZE, MAX_PLAY_SECONDS
from liteai.core import Voice
from snippets.utils import add_callback2gen
from pydub import AudioSegment

from pydub.playback import play


def build_voice(byte_stream: bytes | Iterable[bytes], file_path: str = None):
    if file_path:
        os.makedirs(os.path.dirname(file_path), exist_ok=True)
        if isinstance(byte_stream, bytes):
            logger.debug(f"save voice to {file_path}")
            with open(file_path, "wb") as f:
                f.write(byte_stream)
        else:
            byte_stream = add_callback2gen(byte_stream, dump_voice_stream, path=file_path)
    return Voice(byte_stream=byte_stream, file_path=file_path)


def file2voice(file: str, chunk_size=None):
    def byte_gen():
        with open(file, "rb") as f:
            while True:
                chunk = f.read(chunk_size)
                if not chunk:
                    return
                yield chunk
                logger.debug(f"yielding chunk, with size {len(chunk)}")
    byte_stream = byte_gen()
    return build_voice(byte_stream=byte_stream)


def dump_voice_stream(voice_stream: Iterable[bytes] | bytes, path: str):
    with open(path, "wb") as f:
        logger.debug(f"save voice to {path}")
        if isinstance(voice_stream, bytes):
            f.write(voice_stream)
        else:
            for chunk in voice_stream:
                if chunk is not None and chunk != b'\n':
                    decoded_hex = chunk
                    f.write(decoded_hex)


def get_duration(file_path: str):
    audio = AudioSegment.from_mp3(file_path)
    duration = len(audio) / 1000  # 时长以秒为单位
    return duration


def play_file(file_path: str, max_seconds: int = None):
    logger.debug(f"playing voice from {file_path}")
    with open(file_path, "rb") as f:
        audio_bytes = f.read()
    audio = AudioSegment.from_file(io.BytesIO(audio_bytes), format="mp3")
    if max_seconds:
        audio = audio[:max_seconds * 1000]
    play(audio)


def play_voice(voice: Voice, buffer_size=DEFAULT_VOICE_CHUNK_SIZE, max_seconds=None):

    logger.debug(f"{type(voice.byte_stream)=}, {voice.file_path=}")

    if voice.file_path and os.path.exists(voice.file_path):
        play_file(voice.file_path, max_seconds=max_seconds)
    else:
        logger.debug(f"playing voice from byte stream")
        voice.byte_stream, play_stream = itertools.tee(voice.byte_stream)
        play_bytes(play_stream, buffer_size, max_seconds=max_seconds)


def play_bytes(byte_stream: Iterable[bytes], min_buffer_size=8192*10, max_seconds=None):
    audio_buffer = io.BytesIO()
    producer_finished = threading.Event()
    offset = 0

    def _buf_reader():
        for chunk in byte_stream:
            # logger.debug(f"write {len(chunk)} bytes to buffer")
            audio_buffer.write(chunk)
            audio_buffer.flush()
        producer_finished.set()

    thread = threading.Thread(target=_buf_reader)
    thread.start()
    seconds_left = max_seconds if max_seconds else MAX_PLAY_SECONDS

    def play_segment(check_min_size: bool, seconds_left: int):
        end_offset = audio_buffer.tell()
        raw_size = end_offset - offset
        # logger.debug(f"{producer_finished.is_set()=}, {raw_size=}, {offset=}, {end_offset=}")
        if raw_size > 0:
            # logger.debug(f"{producer_finished.is_set()=}, {raw_size=}, {offset=}, {end_offset=}")
            if not check_min_size or raw_size >= min_buffer_size:
                audio_buffer.seek(offset)
                audio_segment = AudioSegment.from_file(audio_buffer, format="mp3")
                audio_segment = audio_segment[:seconds_left * 1000]
                seconds_left -= len(audio_segment) / 1000
                duration = len(audio_segment)
                logger.debug(f"play audio segment with size:{len(audio_segment.raw_data)}, {raw_size=}, duration:{duration}")
                play(audio_segment)
                return end_offset, seconds_left
        else:
            time.sleep(0.2)
        return offset, seconds_left

    while True:
        if producer_finished.is_set():
            offset, seconds_left = play_segment(False, seconds_left)
            break
        else:
            offset, seconds_left = play_segment(True, seconds_left)
        if seconds_left <= 0:
            break

    # thread.join()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2024/06/25 18:52:00
@Author  :   ChenHao
@Description  :   
@Contact :   jerrychen1990@gmail.com
'''


from typing import Any, Dict, List, Tuple

from litellm import completion

from liteai.core import ModelCard, ModelResponse, Message, ToolDesc, Usage
from liteai.provider.base import BaseProvider
from snippets.utils import add_callback2gen


from litellm import completion

from liteai.utils import acc_chunks, get_text_chunk


# class ZhipuTGILLM(CustomLLM):
#     pass


# zhipu_tgi_llm = ZhipuTGILLM()


# litellm.custom_provider_map = [  # 👈 KEY STEP - REGISTER HANDLER
#     {"provider": "zhipu_tgi", "custom_handler": zhipu_tgi_llm}
# ]


class TGIProvider(BaseProvider):
    key = "tgi"
    allow_kwargs = {"stream", "temperature", "top_p", "max_tokens"}

    def __init__(self, base_url: str = None, **kwargs):
        self.base_url = base_url

    def _support_system(self, model: str):
        return True

    def _parse_zhipu_message(self, messages: List[Message]) -> List[Dict]:
        _input = ""
        for message in messages:
            # logger.debug(f"{message=}")
            _input += f"<|{message['role']}|>\n{message['content']}"
        _input += "<|assistant|>\n"
        messages = [dict(content=_input, role="user")]
        return messages

    def pre_process(self, model: ModelCard, messages: List[Message], tools: List[ToolDesc], stream: bool, **kwargs) -> Tuple[List[dict], dict]:
        messages, tools, kwargs = super().pre_process(model=model, messages=messages, tools=tools, stream=stream, **kwargs)
        if "zhipu" in model or "glm" in model.name:
            messages = self._parse_zhipu_message(messages)
            kwargs.update({"stop":  ["<|endoftext|>", "<|user|>", "<|observation|>"]})
        return messages, tools, kwargs

    def _parse_model(self, model: str) -> str:
        return "huggingface/" + model

    def _inner_complete_(self, model: str, messages: List[dict], stream: bool, tools, ** kwargs) -> Any:
        model = self._parse_model(model)
        api_base = self.base_url + "/generate" if not stream else self.base_url + "/generate_stream"
        response = completion(
            model=model,
            messages=messages,
            api_base=api_base,
            stream=stream,
            **kwargs
        )
        return response

    def post_process(self, response, **kwargs) -> ModelResponse:
        # logger.debug(f"{response=}")
        content = response.choices[0].message.content
        stop = kwargs.get("stop", [])
        for s in stop:
            if s in content:
                content = content[:content.index(s)]
        usage = Usage(**response.usage.model_dump())

        return ModelResponse(content=content, usage=usage)

    def post_process_stream(self, response, **kwargs) -> ModelResponse:
        gen = (e for e in (get_text_chunk(chunk) for chunk in response) if e)
        stops = kwargs.get("stop", [])
        gen = (e for e in gen if e and not e in stops)
        gen = add_callback2gen(gen, acc_chunks)
        return ModelResponse(content=gen)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@Time    :   2024/07/31 15:13:08
@Author  :   ChenHao
@Description  :   openai接口
@Contact :   jerrychen1990@gmail.com
'''


from typing import Any, List, Tuple

from loguru import logger


from liteai.core import ModelCard, ModelResponse, Message, ToolDesc, Usage
from liteai.provider.base import BaseProvider
from openai import OpenAI
from snippets.utils import add_callback2gen

from liteai.utils import image2base64

from snippets import add_callback2gen

from liteai.utils import get_text_chunk, image2base64, acc_chunks


class OpenAIProvider(BaseProvider):
    key: str = "openai"
    allow_kwargs = {"stream", "temperature", "top_p", "max_tokens"}
    api_key_env = "OPENAI_API_KEY"

    def __init__(self, api_key: str = None, base_url: str = None):
        super().__init__(api_key=api_key)
        self.client = OpenAI(api_key=self.api_key, base_url=base_url)

    def pre_process(self, model: ModelCard, messages: List[Message], tools: List[ToolDesc], stream: bool, **kwargs) -> Tuple[List[dict], dict]:
        messages, tools, kwargs = super().pre_process(model=model, messages=messages, tools=tools, stream=stream, **kwargs)
        for message in messages:
            # logger.debug(f"{message=}")
            if message.get("image"):
                base64 = image2base64(message["image"])
                message["content"] = [dict(type="text", text=message["content"]),
                                      dict(type="image_url", image_url=dict(url="data:image/jpeg;base64," + base64))]
                del message["image"]
        return messages, tools, kwargs

    def _inner_complete_(self, model: str, messages: List[dict], stream: bool, tools: List[dict], ** kwargs) -> Any:
        response = self.client.chat.completions.create(
            model=model,
            messages=messages,
            stream=stream,
            **kwargs
        )
        logger.debug(f"{response=}")
        return response

    def post_process(self, response, **kwargs) -> ModelResponse:
        content = response.choices[0].message.content
        usage = Usage(**response.usage.model_dump())
        return ModelResponse(content=content, usage=usage)

    def post_process_stream(self, response, **kwargs) -> ModelResponse:
        gen = (e for e in (get_text_chunk(chunk) for chunk in response) if e)
        gen = add_callback2gen(gen, acc_chunks)
        return ModelResponse(content=gen)


if __name__ == "__main__":
    provider = OpenAIProvider()
    messages = [Message(role="user", content="你好")]
    resp = provider.complete(messages=messages, model="gpt-4o-mini", stream=False)
    print(resp.content)
